"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.clean = exports.getSpan = exports.toString = exports.ready = exports.ensureParsed = exports.tagRegex = exports.parse = exports.getColor = void 0;
const starry_night_1 = require("@wooorm/starry-night");
const dark_style_json_1 = __importDefault(require("./dark-style.json"));
const rules = new Map(Object.entries(dark_style_json_1.default).map(([k, v]) => [k, new Map(Object.entries(v))]));
function getColor(classList) {
    console.assert(classList.length <= 1, `classList too long`);
    const styles = classList.length === 1 ? rules.get(`.${classList[0]}`) : new Map();
    console.assert((styles?.size ?? 0) <= 1, `more styles than just color`);
    const color = styles?.get('color');
    return color;
}
exports.getColor = getColor;
function parse(language, code) {
    if (starryNight == null) {
        throw new Error('you must await ready() to initialize package');
    }
    const scope = starryNight.flagToScope(language);
    if (typeof scope !== 'string') {
        throw new Error(`language ${language} not found`);
    }
    const parsed = starryNight.highlight(code, scope);
    const converted = recurse(parsed);
    return markTree({
        language,
        lines: [],
        chars: converted,
    });
}
exports.parse = parse;
function recurse(node, classes = [], result = []) {
    if (node.type === 'element') {
        console.assert(node.tagName === 'span', `tag was not a span`);
        const className = (node.properties?.className ?? []);
        console.assert(Array.isArray(className), `className was not an array`);
        console.assert(className.length >= 1, `tag did not have a className`);
        console.assert(className.length <= 1, `tag had too many classNames`);
        node.children.forEach((child) => recurse(child, [...classes, className[0]], result));
    }
    else if (node.type === 'root') {
        node.children.forEach((child) => recurse(child, [], result));
    }
    else if (node.type === 'text') {
        for (let i = 0; i < node.value.length; i++) {
            console.assert(classes.length <= 1, `character has too many classes`);
            result.push({
                char: node.value[i],
                classList: classes,
                token: [i, node.value.length],
                sections: [],
            });
        }
    }
    return result;
}
const nextLineRegex = /^\/\/:[^\S\n]*next-line[^\S\n]+([^\n]+?)[^\S\n]*$/;
const thisLineRegex = /^\/\/:[^\S\n]*this-line[^\S\n]+([^\n]+?)[^\S\n]*$/;
const blockStartRegex = /^\/\/<<[^\S\n]*([^\n]+?)[^\S\n]*$/;
const blockEndRegex = /^\/\/>>[^\S\n]*$/;
const sectionStartRegex = /^\/\*<<[^\S\n]*([^\n]+?)[^\S\n]*\*\/$/;
const sectionEndRegex = /^\/\*>>[^\S\n]*\*\/$/;
exports.tagRegex = /^\/\*<[^\S\r\n]*(.*?)[^\S\r\n]*>\*\/$/;
const specialTypes = [
    [nextLineRegex, ['nextLine', 'line']],
    [thisLineRegex, ['thisLine', 'line']],
    [blockStartRegex, ['blockStart', 'line']],
    [blockEndRegex, ['blockEnd', 'line']],
    [sectionStartRegex, ['sectionStart', 'span']],
    [sectionEndRegex, ['sectionEnd', 'span']],
    [exports.tagRegex, ['tag', 'span']],
];
function getSpecialType(span) {
    for (const [regex, result] of specialTypes) {
        if (regex.test(span))
            return result;
    }
    return [];
}
function* spans(chars) {
    let i = 0;
    while (i < chars.length) {
        const char = chars[i];
        console.assert(char.token[0] === 0, `token was not beginning of span`);
        yield [chars.slice(i, i + char.token[1]), i];
        i += char.token[1];
    }
}
function markTree(parsed) {
    const chars = parsed.chars;
    const lineCount = 1 +
        chars.reduce((prior, { char }) => {
            return prior + (char === '\n' ? 1 : 0);
        }, 0);
    const lines = new Array(lineCount)
        .fill(1)
        .map(() => ({ tags: [] }));
    let ln = 0;
    const blocks = [];
    const sections = [];
    let sectionHold = null;
    const result = [];
    for (let i = 0; i < chars.length; i++) {
        const char = chars[i];
        if (char.token[0] === 0 && sectionHold != null) {
            sections.unshift(sectionHold);
            sectionHold = null;
        }
        result.push({
            ...char,
            sections: [...sections],
        });
        if (char.token[0] !== 0) {
            continue;
        }
        const span = getSpan(chars, i);
        if (char.char === '\n') {
            ln++;
            lines[ln].tags.push(...blocks);
        }
        else if (char.classList.length === 1 && char.classList[0] === 'pl-c') {
            if (nextLineRegex.test(span)) {
                const name = span.match(nextLineRegex)?.[1];
                lines[ln + 1].tags.push(name);
            }
            else if (thisLineRegex.test(span)) {
                const name = span.match(thisLineRegex)?.[1];
                lines[ln].tags.push(name);
            }
            else if (blockStartRegex.test(span)) {
                blocks.push(span.match(blockStartRegex)?.[1]);
            }
            else if (blockEndRegex.test(span)) {
                blocks.pop();
            }
            else if (sectionStartRegex.test(span)) {
                sectionHold = span.match(sectionStartRegex)?.[1];
            }
            else if (sectionEndRegex.test(span)) {
                sections.shift();
            }
        }
    }
    return {
        ...parsed,
        lines,
        chars: result,
    };
}
/**
 * @internal
 * @param input - parsed or parsable variable
 * @returns parsed code
 */
function ensureParsed(input) {
    if (Array.isArray(input)) {
        return parse(input[0], input[1]);
    }
    else if (typeof input === 'object' && 'code' in input) {
        return parse(input.lang, input.code);
    }
    else {
        return input;
    }
}
exports.ensureParsed = ensureParsed;
let starryNight = null;
const starryNightPromise = (0, starry_night_1.createStarryNight)(starry_night_1.all);
starryNightPromise.then((sn) => (starryNight = sn));
function ready() {
    return starryNightPromise;
}
exports.ready = ready;
function toString(code) {
    const parsed = ensureParsed(code);
    const result = [];
    parsed.chars.forEach(({ char }) => result.push(char));
    return result.join('');
}
exports.toString = toString;
function getSpan(tree, at) {
    const [back, length] = tree[at].token;
    const start = at - back;
    const end = start + length;
    let result = '';
    for (let i = start; i < end; i++) {
        result += tree[i].char;
    }
    return result;
}
exports.getSpan = getSpan;
/**
 * Removes all code-fns tags.
 *
 * @param code - the parsed or parsable code to clean
 */
function clean(code) {
    const parsed = ensureParsed(code);
    const result = [];
    let lineNumber = 0;
    let finalNumber = 1;
    let lineSkip = false;
    const lines = [];
    for (const [span] of spans(parsed.chars)) {
        if (span.length === 1 && span[0].char === '\n')
            lineNumber++;
        if (lineSkip) {
            console.assert(span.length === 1 && span[0].char === '\n', `expected a new line`);
            lineSkip = false;
            continue;
        }
        else if (span.length === 1 && span[0].char === '\n') {
            lines.push({
                ...parsed.lines[lineNumber],
                number: finalNumber,
            });
            finalNumber++;
        }
        console.assert(span[0].classList.length === 1, `multiple classes found`);
        const specialTypes = getSpecialType(span.reduce((text, { char }) => text + char, ''));
        if (span[0].classList[0] !== 'pl-c' || specialTypes.length === 0) {
            result.push(...span);
        }
        else if (specialTypes.includes('line')) {
            lineSkip = true;
        }
    }
    lines.push({
        ...parsed.lines[lineNumber],
        number: finalNumber,
    });
    if (lineSkip) {
        lines.pop();
        console.assert(result[result.length - 1].char === '\n', `expected a new line`);
        result.pop();
    }
    return {
        ...parsed,
        chars: result,
        lines,
    };
}
exports.clean = clean;
